#include <math.h>
#include "Distance.h"
#include "Utils.h"


#define INC_BUFFER_DIST 1000
#define SIZE_BUFFER_TypeFloat 100
#define INC_BUFFER_ANC 10

TypeFloat computeVariance(TypeDistance dist) {	
	TypeNumber i, j;
	TypeFloat mean= 0.0, var = 0.0;
	for(i=1; i<dist.number; i++)
		for(j=0; j<i; j++)
			mean += dist.table[(i*(i-1))/2+j];
			
	for(i=1; i<dist.number; i++)
		for(j=0; j<i; j++)
			var += pow(2, dist.table[(i*(i-1))/2+j]-mean);
	return var;
}

void printDistanceRaw(FILE *f, TypeDistance dist) {
	TypeNumber i, j;
	for(i=1; i<dist.number; i++) {
		fprintf(f, "'%s'", dist.name[i]);
		for(j=0; j<i; j++)
			fprintf(f, "\t%f",dist.table[(i*(i-1))/2+j]);
		fprintf(f, "\n");
	}
}

void printDistancePhylip(FILE *f, TypeDistance dist) {
	TypeNumber i, j;
	fprintf(f, "%d\n", dist.number);
	for(i=0; i<dist.number; i++) {
		fprintf(f, "'%s'", dist.name[i]);
		for(j=0; j<i; j++)
			fprintf(f, "\t%f",dist.table[(i*(i-1))/2+j]);
		fprintf(f, "\t0.0\n");
		for(j=i+1; j<dist.number; j++)
			fprintf(f, "\t%f",dist.table[(j*(j-1))/2+i]);
		fprintf(f, "\n");
	}
}


void printDistanceNexus(FILE *f, TypeDistance dist) {
	TypeNumber i, j;
	printPreambleNexus(f, dist.name, dist.number);
	fprintf(f, "BEGIN distances;\n");
	fprintf(f, "    DIMENSIONS ntax=%d;\n", dist.number);
	fprintf(f, "    FORMAT\ntriangle=LOWER\ndiagonal\nlabels\nmissing=?\n;\n");
	fprintf(f, "MATRIX\n");
	for(i=0; i<dist.number; i++) {
		fprintf(f, "'%s'", dist.name[i]);
		for(j=0; j<i; j++)
			fprintf(f, "\t%f",dist.table[(i*(i-1))/2+j]);
		fprintf(f, "\t0.0\n");
	}
	fprintf(f, ";\n");
	fprintf(f, "END;\n");
}



TypeDistance computeDistance(TypeSetOfSequences set) {
	TypeDistance dist;
	TypeNumber i, j;
	TypePosition **occ, n, p;
	
	occ = (TypePosition**) monmalloc(set.number*sizeof(TypePosition*));
	for(n=0; n<set.number; n++) {
		occ[n] = (TypePosition*) monmalloc(set.cardinal*sizeof(TypePosition));
		for(p=0; p<set.cardinal; p++)
			occ[n][p] = 0;
		for(p=0; p<set.size[n]; p++)
			occ[n][set.sequence[n][p]]++;
	}
	dist.number = set.number;
	dist.name = set.name;
	dist.table = (TypeFloat*) monmalloc(((dist.number*(dist.number-1))/2)*sizeof(TypeFloat));
	for(i=1; i<dist.number; i++)
		for(j=0; j<i; j++) {
			TypeNumber ind = (i*(i-1))/2+j;
			TypeFloat sizeMin = (TypeFloat) MIN(set.size[i], set.size[j]);
			dist.table[ind] = 0.0;
			for(p=0; p<set.cardinal; p++)
				dist.table[ind] += ((TypeFloat) MIN(occ[i][p], occ[j][p]));
			dist.table[ind] = 1-dist.table[ind]/sizeMin;
		}
	for(n=0; n<set.number; n++)
		monfree((void*) occ[n]);
	monfree((void*) occ);
	return dist;
}

TypeDistance readDistances(FILE *f) {
	char c, *bufferTypeFloat;
	TypeNumber sizeTmp = 0, sizeBufferDist;
	TypeFloat *bufferDist;
	TypeDistance dist;

	sizeBufferDist = INC_BUFFER_DIST;
	bufferDist = (TypeFloat *) monmalloc(sizeBufferDist*sizeof(TypeFloat));
	bufferTypeFloat = (char *) monmalloc(SIZE_BUFFER_TypeFloat*sizeof(char));
	c = getc(f);
	do {
		int indTmp = 0;
		TypeFloat distTmp;
		while(c != EOF && IsSeparator(c)) {
			c = getc(f);
		}
		while(c != EOF && indTmp<SIZE_BUFFER_TypeFloat  && !IsSeparator(c)) {
			if(c != ',')
				bufferTypeFloat[indTmp++] = c;
			else
				bufferTypeFloat[indTmp++] = '.';
			c = getc(f);
		}
		if(indTmp<SIZE_BUFFER_TypeFloat)
			bufferTypeFloat[indTmp++] = ' ';
		if(indTmp>1 && sscanf(bufferTypeFloat, "%f", &distTmp) == 1) {
			if(sizeTmp >= sizeBufferDist) {
				sizeBufferDist += INC_BUFFER_DIST;
				bufferDist = (TypeFloat *) monrealloc((void *) bufferDist, sizeBufferDist*sizeof(TypeFloat));
			}
			bufferDist[sizeTmp++] = distTmp;
		}
	} while(c != EOF );
	free((void*) bufferTypeFloat);
	dist.number = (TypeNumber) floor((-1+sqrt(8*sizeTmp+1))/2);
	dist.table = (TypeFloat*) monrealloc((void *) bufferDist, ((dist.number*(dist.number-1))/2)*sizeof(TypeFloat));
	return dist;
}


TypeFloat distance(TypeDistance d, TypeNumber i, TypeNumber j) {
	if(i == j || i>=d.number || j>=d.number || i<0 || j<0)
		return 0;
	if(i<j)
		return d.table[(j*(j-1))/2+i];
	else
		return d.table[(i*(i-1))/2+j];
}

TypeFloat distanceClassesMin(TypeDistance d, TypeNumber *tab1, TypeNumber n1, TypeNumber *tab2, TypeNumber n2) {
	TypeNumber i, j;
	TypeFloat min;
	if(n1 == 0 || n2 == 0)
		return 0.0;
	min = distance(d, tab1[0], tab2[0]);
	for(i=0; i<n1; i++)
		for(j=0; j<n2; j++)
			if(min>distance(d, tab1[i], tab2[j]))
				min = distance(d, tab1[i], tab2[j]);
	return min;
}

TypeFloat distanceClassesMoy(TypeDistance d, TypeNumber *tab1, TypeNumber n1, TypeNumber *tab2, TypeNumber n2) {
	TypeNumber i, j;
	TypeFloat moy;
	if(n1 == 0 || n2 == 0)
		return 0.0;
	moy = 0.0;
	for(i=0; i<n1; i++)
		for(j=0; j<n2; j++)
				moy += distance(d, tab1[i], tab2[j]);
	return moy/(n1*n2);
}






