#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "Sequences.h"
#include "Distance.h"
#include "Decode.h"
#include "Neighbor.h"
#include "Utils.h"


#define SIZE_BUFFER_CHAR 300

#define HELPMESSAGE "\nusage: locdec [options] <order> <input file> <output file>\n\nThe input file has to be in Fasta format.\n\nOptions:\n\n-h                      output help\n\n-f <f, s, m, x>         select the format of output for distances matrix \n                        (r: raw, p: phylip, n: nexus)\n\n-s <d, r, p>            select the type of alphabet (d: DNA, r: RNA, p: Protein)\n                        maybe useful to handle ambiguity characters\n\n-t <p, m, a>            select the type of output:\n                         * l: (default) return the local decoding of order <order> in Fasta format\n                         * d: return corresponding matrix of distances,\n			      format depending on option 'o' (raw by default) \n"
#define AMBIMESSAGE "The ambiguity codes are for DNA and RNA:\n                                         * Y: T/U or C\n                                         * R: A or G\n                                         * M: A or C\n                                         * K: T/U or G\n                                         * W: T/U or A\n                                         * S: C or G\n                                         * B: T/U, C, or G\n                                         * D: T/U, A, or G\n                                         * H: T/U, C, or A\n                                         * V: C, A, or G\n                                         * N: T/U, C, A, or G\n\nand for protein:\n                                         * B: D or N\n                                         * Q or E\n                                         * X: Unidentified\n"

int main(int argc, char **argv) {		
	TypePosition order;
	char option[256], inputFileName[SIZE_BUFFER_CHAR], outputFileName[SIZE_BUFFER_CHAR], bufferOutput[SIZE_BUFFER_CHAR], *table, 
	outputFormat = 'r', typeTask = 'c', typeAlphabet = '?';
	TypeSetOfSequences set;
	FILE *fi, *fo;
	int i = 1;

	for(i=0; i<256; i++)
		option[i] = 0;
	for(i=1; i<argc && *(argv[i]) == '-'; i++) {
		int j;
		for(j=1; argv[i][j] != '\0'; j++)
			option[argv[i][j]] = 1;
		if(option['f']) {
			option['f'] = 0;
			if((i+1)<argc && sscanf(argv[i+1], "%c", &outputFormat) == 1)
				i++;
		}
		if(option['s']) {
			option['s'] = 0;
			if((i+1)<argc && sscanf(argv[i+1], "%c", &typeAlphabet) == 1)
				i++;
		}
		if(option['t']) {
			option['t'] = 0;
			if((i+1)<argc && sscanf(argv[i+1], "%c", &typeTask) == 1)
				i++;
		}
		if(option['h']) {
			printf("%s\n", HELPMESSAGE);
			exitProg(ExitOk, NULL);
		}
	}
	if (i>=argc || sscanf(argv[i++], "%ld", &order) != 1) exitProg(ErrorArgument, HELPMESSAGE);
	if (i>=argc || sscanf(argv[i++], "%s", inputFileName) != 1) exitProg(ErrorArgument, HELPMESSAGE);
	if (i>=argc || sscanf(argv[i++], "%s", outputFileName) != 1) exitProg(ErrorArgument, HELPMESSAGE);

	switch(typeAlphabet) {
		case 'd':
			table = (char*) monmalloc((strlen(DNA)+1)*sizeof(char));
			strcpy(table, DNA);
			break;
		case 'r':
			table = (char*) monmalloc((strlen(RNA)+1)*sizeof(char));
			strcpy(table, RNA);
			break;
		case 'p':
			table = (char*) monmalloc((strlen(PRO)+1)*sizeof(char));
			strcpy(table, PRO);
			break;
		case '?':
		default:
			table = (char*) monmalloc(sizeof(char));
			table[0] = '\0';
	}
	if(fi = fopen(inputFileName, "r")) {
		set = readSequencesFasta(fi, table, typeAlphabet == '?');
		fclose(fi);
	} else {
		exitProg(ErrorReading, inputFileName);
	}
	if(fo = fopen(outputFileName, "w")) {
		switch(typeTask) {
			case 'd':
				switch(outputFormat) {
					case 'r':
						printDistanceRaw(fo, computeDistance(localDecode(set, order)));
						break;
					case 'p':
						printDistancePhylip(fo, computeDistance(localDecode(set, order)));
						break;
					case 'n':
						printDistanceNexus(fo, computeDistance(localDecode(set, order)));
				}
				break;
/*			case 't':
				if(outputFormat == 'r')
					printTreeNewick(fo, computeTree(computeDistance(localDecode(set, order))));
				else
					printTreeNexus(fo, computeTree(computeDistance(localDecode(set, order))));
				break;
*/			case 'l':
			default:
				printDecodedSequencesFasta(fo, set, localDecode(set, order), 50);
				break;
		}
		fclose(fo);
	} else {
		exitProg(ErrorWriting, outputFileName);
	}
	exitProg(ExitOk,"Bye");
	return 0;
}
