#ifndef DecodeF
#define DecodeF

#include <stdlib.h>
#include <stdio.h>
#include "Sequences.h"

typedef struct SUFFIXNODE {
	struct SUFFIXNODE *trans, *next, *suffix;
	TypePosition start, end;
} TypeSuffixNode;

typedef struct SUFFIXTREE {
	TypeSuffixNode *root, *bottom;
	TypeSymbol *sequence;
} TypeSuffixTree;

typedef struct LINK {
	struct LINK *next;
	TypePosition position;
} TypeLink;

typedef struct LOCALNODE {
	struct LOCALNODE *trans, *next, *ancestor, *nlink, *ptrans, *pnext;
	int state;
	TypePosition position;
} TypeLocalNode;

typedef struct LOCALTREE {
	TypeLocalNode **leaf;
	TypePosition size;
} TypeLocalTree;

TypeSetOfSequences localDecode(TypeSetOfSequences set,  TypePosition order);
TypeOneSequence localDecodeOneSequence(TypeOneSequence seq, TypePosition order);

void printDecodedSequencesFasta(FILE *f, TypeSetOfSequences set, TypeSetOfSequences dec, int sizeLine);

void freeSuffixTree(TypeSuffixTree *suffixTree);
void freeLocalTree(TypeLocalTree *localTree);
void printSuffixTree(FILE *f, TypeSuffixTree *suffixTree);
void printLocalTree(FILE *f, TypeLocalTree *localTree);
void printSequenceDebug(FILE *f, TypeOneSequence s, int sizeLine);

#endif

#ifndef DistancesF
#define DistancesF

#include <stdlib.h>
#include <stdio.h>
#include "Sequences.h"
typedef float TypeFloat;

typedef struct DISTANCES {
	char **name;
	TypeNumber number;
	TypeFloat *table;
} TypeDistance;


TypeDistance readDistances(FILE *f);
TypeDistance computeDistance(TypeSetOfSequences set);
TypeFloat computeVariance(TypeDistance dist);
TypeFloat distance(TypeDistance d, TypeNumber i, TypeNumber j);
TypeFloat distanceClassesMin(TypeDistance d, TypeNumber *tab1, TypeNumber n1, TypeNumber *tab2, TypeNumber n2);
TypeFloat distanceClassesMoy(TypeDistance d, TypeNumber *tab1, TypeNumber n1, TypeNumber *tab2, TypeNumber n2);
void printDistanceRaw(FILE *f, TypeDistance dist);
void printDistanceNexus(FILE *f, TypeDistance dist);
void printDistancePhylip(FILE *f, TypeDistance dist);

#endif
#ifndef DendrogramF
#define DendrogramF

#include <stdlib.h>
#include <stdio.h>
#include "Distance.h"

typedef struct TREENODE {
	struct TREENODE *child1, *child2, *child3;
	TypeFloat length;
	TypeNumber ident;
} TypeTreeNode;

typedef struct NEIGHBORTREE {
	TypeTreeNode *root;
	char **name;
	TypeNumber number;
} TypeNeighborTree;

TypeNeighborTree computeTree(TypeDistance dist);

void printTreeNewick(FILE *f, TypeNeighborTree tree);
void printTreeNexus(FILE *f, TypeNeighborTree tree);

#endif

#ifndef SequencesF
#define SequencesF

#include <stdlib.h>
#include <stdio.h>

typedef /*unsigned*/ long TypePosition;
typedef unsigned int TypeSymbol;
typedef unsigned int TypeNumber;
/*Structure storing a set of sequence*/
typedef struct AMBIGUITY {
	char *car;
	TypeSymbol **set, *size, number;
} TypeAmbiguity;

/*Structure storing one sequence*/
typedef struct ONESEQUENCE {
/*cardinal of the alphabet*/
	TypePosition size;
	TypeSymbol cardinal;
	TypeSymbol *sequence;
	TypeAmbiguity ambiguity;
} TypeOneSequence;

/*Structure storing a set of sequence*/
typedef struct SEQUENCES {
	char **name, *table, *title;
/*number of TypeSetOfSequences include in the set*/
	TypeNumber number;
/*cardinal of the alphabet*/
	TypeSymbol cardinal;
/*size[i] size of the sequence i*/
	TypePosition *size;
/*table of all the sequences*/
	TypeSymbol **sequence;
	TypeAmbiguity ambiguity;
} TypeSetOfSequences;

TypeOneSequence toOne(TypeSetOfSequences set);
void printSequencesFasta(FILE *f, TypeSetOfSequences s, int sizeLine);
TypeSetOfSequences readSequencesFasta(FILE *f, char *table, int canInc);
TypeSetOfSequences readSequencesRaw(FILE *f);
void writeSequences(FILE *f, TypeSetOfSequences seq);
void freeSequences(TypeSetOfSequences seq);
TypeAmbiguity getXNAAmbiguity();
TypeAmbiguity getProteinAmbiguity();
TypeSetOfSequences clone(TypeSetOfSequences s);

#define SEP -1
#define FIN -1
#define DNA "ACGTYRMKWSBDHVN" /*"ACGT" + "YRMKWSBDHVN"*/
#define RNA "ACGUYRMKWSBDHVN" /*"ACGU" + "YRMKWSBDHVN"*/
#define PRO "DEGNQCSTYAVLIPFMWKRHBZX" /*"DEGNQCSTYAVLIPFMWKRH" + "X"*/

#endif
#ifndef UtilsF
#define UtilsF
#include <stdlib.h>
#include <stdio.h>

typedef enum TS {
	ErrorArgument=0,
	ErrorInit,
	ErrorReading,
	ErrorWriting,
	ErrorMemory,
	ErrorExec,
	ExitOk
} TypeExit;

void printPreambleNexus(FILE *f, char **name, int number);
void exitProg(TypeExit code, char *message);
void *monmalloc(long size);
void *monrealloc(void *in, long size);
int IsSeparator(char c);
int IsItemSeparator(char c);
int IsLineSeparator(char c);
int readLine(FILE *f, char *buffer);
int readItem(FILE *f, char *buffer);
int tokenize(char *src, char **dest);
int find(char *src, char **dest, int size);
void fixSpace(char *src);
char *truncFileName(char* name);
char *getExtension(char* name);

#define MAX(x,y) ((x)>(y)?(x):(y))
#define MIN(x,y) ((x)<(y)?(x):(y))
#define ABS(x) ((x)<(0)?(-x):(x))
#define NEG_INFTY -9.999e99
#endif
