#! /usr/bin/env python

# -*- coding: iso-8859-1 -*-

##
##  Copyright (C) 2010 E. Corel, G. Didier, C. Devauchelle, G. Grasseau
##
##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or (at
##  your option) any later version.
## 
##  This program is distributed in the hope that it will be useful, but
##  WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
## 
##  You should have received a copy of the GNU General Public License
##  along with this program; if not, write to the Free Software
##  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
## 
##

import sys, getopt, os, time, copy
#from altgraph import Graph
# A voir ???
import mySyntax

##
# Anchors
#

class AnchorSet:

    def __init__(self, dict_=None ):

        ## Anchor list.
        self.anchors = {};

        ## Classes list
        self.classes = {};


        if dict_:
            dict_.values().sort( mySyntax.lexicoTuple )
            self.anchors = dict_ 
            self.classes = self.getClasses()

    #======= Constructor methods

    def getClasses(self):
        """ Computes the inverse dictionary of anchors : site -> name of anchor -- ignores singletons; moreover can be sped up with tG.invertDico1 if needed """
        classes = {}
        ## GG ??? Ecriture bizarre
        for node in (node for node in self.anchors.keys()):
            ## Singletons are ignored
            if len(self.anchors[node]) != 1 : 
                for i in self.anchors[node]:
                    classes[i] = node

        return classes


    ## Write the anchor list in a FASTA file.
    #  @param name file name.
    #  @param ext file name ssuffix.
    #  @param seq_set is the  SequenceSet object proceeded 
    #
    ## Write the anchor list in a FASTA formated file. Anchors sites are replaced in the input 
    # FASTA file by the anchor name (ex . A C G T A_322 C C T G).

    def writeAnchors( self, name, ext=None, SeqSet=None, 
                        lineLength=80):
        mark='*'

        ## GG ???
        smin=1

        if ( (SeqSet != None) and ( len( self.anchors.keys() ) != 0) )  :

            SeqSet_out = copy.deepcopy( SeqSet )

            for n in self.anchors.keys():
# treatment of anchors which are equal to a leaf : 
# replace the leaf identifier by its letter+'*'
                if len(self.anchors[n]) <= smin:           
                    seq, pos = n
                    # Mark leaves
                    SeqSet_out.sites[seq][pos] = SeqSet_out.sites[seq][pos]+mark        
                else:                                            
                    # Treatment of anchors which are in inner nodes
                    for site in self.anchors[n]:
                        seq,pos = site[0],site[1]
                        # modifies in place the FastaObject by replacing letters 
                        # in original sites by names of nodes in hybrid partition
                        SeqSet_out.sites[seq][pos] = n    
            # Save the new FASTA file
            SeqSet_out.writeFastaFile(name=name,ext=ext,lineLength=lineLength)  
    

    ## Write the anchor list.
    #  @param name file name.
    #  @param dir directory.
    #  @param ext file name ssuffix.
    #  @param seq_set is the  SequenceSet object proceeded 
    #  @param PosMode write with the position format 
    # Write the anchor list in the file "name". The output format is the folowing : 
    # [A1A2A3...An] [B1B2..Bm] ...  where Ak, Bl are the???
    def writeWindows( self, name=None, ext=None, SeqSet=None, PosMode=False):
        """ Writes the store file for anchors --window format """

        if not isinstance(name, str): 
            sys.exit("writeWindows: No file name")   

        if isinstance(ext, str): 
            name = name + ext

        f = open(name,'w')

        if( SeqSet != None ):
        
             for i,v in self.anchors.iteritems():
                try:
                    n = int(i.split('_')[1])
                    # print n
                except:
                    pass
                if len(v) != 1:
                    # ??? A voir plutard
                    v.sort(mySyntax.lexicoTuple)
                    try:
                        # Get the set of letters which occurs in the 
                        # site list which belon to the anchor
                        # ??? letters = self.letterAnchor(v)
                        letters = SeqSet.getLettersOfSites( v ) 
                        f.write('>Anchor'+str(i)+' '+letters+'\n')
                        if ( PosMode ):
                            for site in v:
                                seg = SeqSet.getNeighbouring(site,n,sep='_')
                                print >> f, "%-15s %s" % (str(site),seg)
                                print >>f
                        else:
                            for site in v:
                                seg = SeqSet.getNeighbouring(site,n,sep='')
                                print >> f, "%s" % (seg)
                    except TypeError:
                        pass
        f.close()
                    




# GG ??? a mettre dans ComputeDistances
    def computeMatrix(self,dim):
        """ Computes similarity matrix (without normalisation) for anchors """
        m = mySyntax.initMatrix(dim)
        for c in self.anchors.keys():
            seq = map(lambda x:x[0],self.anchors[c])  # the list of sequences where anchor c appears
            seqset = list(set(seq))                                      # the same, without repetitions
            seqset.sort(key=lambda x:int(x),cmp=lambda x,y:x-y) # sorted list
            counts = dict([(s,seq.count(s)) for s in seqset])  # the count of times where anchor c appears in sequence s
            k = len(seqset)                                           # the number of different sequences where anchor c appears
            for i in range(k):
                for j in range(k)[i+1:]:
                    m[seqset[i]][seqset[j]] += min(counts[seqset[i]],counts[seqset[j]])
        return m

    def normaliseMatrix(self,m):
        lengths = dict([(i,len(self.sites[i])) for i in self.sites.keys()]) # lengths of the sequences
        k = len(self.sequences)
        for i in range(k):
            for j in range(k)[i+1:]:
                m[i][j] = 1-float(m[i][j])/float(min(lengths[i],lengths[j]))
        for i in range(k):
            for j in range(k)[:i]:
                m[i][j] = m[j][i]
        return m

    def outputMatrix(self,outname,format,ext=None):
        dim = len(self.sequences)
        m = self.computeMatrix(dim)
        m = self.normaliseMatrix(m)
        mySyntax.writeNexus(outname,m,self.sequences,format,comment=outname,ext=ext)
