#! /usr/bin/env python

# -*- coding: iso-8859-1 -*-

##
##  Copyright (C) 2010 E. Corel, G. Didier, C. Devauchelle, G. Grasseau
##
##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or (at
##  your option) any later version.
## 
##  This program is distributed in the hope that it will be useful, but
##  WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
## 
##  You should have received a copy of the GNU General Public License
##  along with this program; if not, write to the Free Software
##  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
## 
##

## @package BuildEquivClasses
#  Short documentation (sum up).
#
#  More details ???.


import SequenceSet,STree, EquivClassesLayers
import MS4_lib
import re, sys,os,itertools,time
from math import ceil
import altgraph
#import boost.graph as bgl
# import networkx as NX
from altgraph import Graph, GraphAlgo, Dot
from collections import deque
import string
# GG ???
TIMETEST = 0

## Abstract Class to build EquivClasses.
#
#  More details ???.
class BuildEquivClasses:

    ## Virtual method to .
    #  @param self the object pointer.
    #  @param seqset_ a SequenceSet object.
    #
    #  More details ???.
    def build( self, seqset_=None):
        raise NotImplementedError

## Class to build EquivClasses using exact NLD method.
#
#  More details ???.
class BuildExactEquivClasses(BuildEquivClasses):
    
    ## Minimum order value.
    order_min = 0;

    ## Minimum order value.
    order_max = 0;

    ## Implementaion mode.
    mode = None;
    
    ## Constructor .
    #  @param self the object pointer.
    #  @param seqset_ a SequenceSet object.
    #  @param order_l order range.
    #  @param mode selects different implementation (None, 'SuffixTree'
    #         'NLDEquivClasses') 
    #
    # Compute the equivalent classes for different orders in the range 
    # 'order_l' :
    #       order_l = [order_min, order_max] > 0 
    #                 with (order_min < order_max ) is a range of orders
    #       order_l = [order_min, ... ] > 0 is a fixed order value.
    #       order_l = [order_min, ... ] < 0 0 is an order list determined
    #                 by the suffix tree.
    #
    #  Three implementation are provided:
    #    mode = None (default), chouses automatically the best implementation.
    #    mode = 'SuffixTree', the EquivClasses are computed in the python code.
    #    mode = 'NLDEquivClasses', the EquivClasses are not computed in 
    #           the python code.
    def __init__( self, order_l = [-1,-1],  mode = None ):

        # GG ??? test args order
        if( (order_l[0] > 0) ) :
          self.order_min = order_l[0]
          if( (order_l[1] >= order_l[0]) ) :
              self.order_max = order_l[1]
          else :
              self.order_max = order_l[0]
        else :
            self.order_min = -1
            self.order_max = -1
            

        # GG ??? test in 'SuffixTree', ...
        if ( mode != None ) :
            self.mode = mode 
        else :
            self.mode = 'NLDEquivClasses'
 

    def build( self, seqset_ ):
        """ Return an EquivClasses object """
        # GG ??? Filter order list with the suffix tree
        # st = STree.SuffixTree(seqset_.filename,option='sites')
        st = STree.SuffixTree(seqset_, option='sites') 
        nlist = st.signifValues()
        sys.stderr.write(str(nlist) +'\n')
        # ??? A valider nList = nlist[self.order_min: self.order_max]
        nList = []
        for i in range( len(nlist) ) :
            if ( (nlist[i] >= self.order_min) 
                 and  (nlist[i] <= self.order_max) ) :
                nList.append(nlist[i]) 

        nList.reverse()
        nList.sort(cmp=lambda x,y:y-x)
        # sys.stderr.write(str(time.clock()-t0)+'\n')
        # sys.stderr.write('iteration over values '+str(nList)+'\n')
        sys.stderr.write('Building Equivalences Classes\n') 
         
        ECLayers = EquivClassesLayers.EquivClassesLayers( seqset_.sites )
        for n in nList:
            # GG ??? d = Decodable.Decodable(fname=self.filename,n=n,optDec='decode',optSuff=None)      # constructs the decoding of filename at order n
            # dico = d.letterLocation()                       # dict GDclass -> (sequence, position)
            # ???  Utiliser les PTree plustot que les EquivClasses
            # .addLayer Doit se faire dans les EquivClasses
            dec_seqset_ = self.getExactDecodedSeq( seqset_, n )
            # Ecriture confuse (ecrire explicitement la boucle
            # Inutile ???
            # dic = dict( [( seq, dec_seqset_.SeqNames.index(seq)) 
            #             for seq in dec_seqset_.SeqNames] )
            ECLayers.addLayer( n,  dec_seqset_.getECLocation( )  )
            sys.stderr.write(str(n)+' ')

        sys.stderr.write('\n')
        return ECLayers

    def getExactDecodedSeq( self, SeqSet, order):
        """ Invalide call C ???
        fname     = seqset_.filename 
        out_fname = '/tmp/' + fname+str(order)+'.dec'
        if order :        
            cmd = 'decode ' + str(order) + ' ' + fname  + ' ' + out_fname
            os.system(cmd)

            # ??? suffixe  suf = '_'+str(n)
            # ??? Phantom
        """
        # Build a String List Argument
        seqstrings=[]
        for seq in SeqSet.sites:
            str = ""
            for i in SeqSet.sites[seq]:
                str = str + i
            seqstrings.append( str )
                
        # sys.stderr.write('in NLDExact \n')
        StringList = MS4_lib.NLDExact( seqstrings, order )
        # sys.stderr.write('out NLDExact \n')
        
        out_seqset = SequenceSet.SequenceSet( SeqSet=SeqSet,
                                              StringList=StringList,
                                              phantom=1)
        
        return out_seqset
       

## Class to build EquivClasses using exact NLD method.
#
#  More details ???.
class BuildMismatchEquivClasses(BuildEquivClasses):
    
    ## Minimum order value.
    order_min = 0;

    ## Minimum order value.
    order_max = 0;

    ## Implementaion mode.
    mode = None;
    
    ## Constructor .
    #  @param self the object pointer.
    #  @param seqset_ a SequenceSet object.
    #  @param order_l order range.
    #  @param mode selects different implementation (None, 'SuffixTree'
    #         'NLDEquivClasses') 
    #
    # Compute the equivalent classes for different orders in the range 
    # 'order_l' :
    #       order_l = [order_min, order_max] > 0 
    #                 with (order_min < order_max ) is a range of orders
    #       order_l = [order_min, ... ] > 0 is a fixed order value.
    #       order_l = [order_min, ... ] < 0 0 is an order list determined
    #                 by the suffix tree.
    #
    #  Three implementation are provided:
    #    mode = None (default), chouses automatically the best implementation.
    #    mode = 'SuffixTree', the EquivClasses are computed in the python code.
    #    mode = 'NLDEquivClasses', the EquivClasses are not computed in 
    #           the python code.
    def __init__( self, order_l = [-1,-1],  mode = None ):

        # GG ??? test args order
        if( (order_l[0] > 0) ) :
          self.order_min = order_l[0]
          if( (order_l[1] >= order_l[0]) ) :
              self.order_max = order_l[1]
          else :
              self.order_max = order_l[0]
        else :
            self.order_min = -1
            self.order_max = -1
            

        # GG ??? test in 'SuffixTree', ...
        if ( mode != None ) :
            self.mode = mode 
        else :
            self.mode = 'NLDEquivClasses'
 

    def build( self, seqset_ ):
        """ Return an EquivClasses object """
        # Obtain the minimal list of orders to run
        st = STree.SuffixTree(seqset_, option='sites') 
        nlist = st.signifValues()
        sys.stderr.write(str(nlist) +'\n')
        # ??? A valider nList = nlist[self.order_min: self.order_max]
        nList = []
        for i in range( len(nlist) ) :
            if ( (nlist[i] >= self.order_min) 
                 and  (nlist[i] <= self.order_max) ) :
                nList.append(nlist[i]) 

        nList.reverse()
        nList.sort(cmp=lambda x,y:y-x)
        # sys.stderr.write(str(time.clock()-t0)+'\n')
        # sys.stderr.write('iteration over values '+str(nList)+'\n')
        sys.stderr.write('Building Equivalences Classes\n')
          
        ECLayers = EquivClassesLayers.EquivClassesLayers( seqset_.sites )
        for n in nList:
            # GG ??? d = Decodable.Decodable(fname=self.filename,n=n,optDec='decode',optSuff=None)      # constructs the decoding of filename at order n
            # dico = d.letterLocation()                       # dict GDclass -> (sequence, position)
            # ???  Utiliser les PTree plustot que les EquivClasses
            # .addLayer Doit se faire dans les EquivClasses
            dec_seqset_ = self.getMismatchDecodedSeq( seqset_, n )
            # Ecriture confuse (ecrire explicitement la boucle
            # dic = dict( [( seq, dec_seqset_.SeqNames.index(seq)) 
            #             for seq in dec_seqset_.SeqNames] )
            ECLayers.addLayer( n,  dec_seqset_.getECLocation( )  )
            sys.stderr.write(str(n)+' ')

        sys.stderr.write('\n')
        return ECLayers

    def getMismatchDecodedSeq( self, seqset_, order):
        """ Invalid ??? - System call
        fname     = seqset_.filename 
        ## ???
        out_fname = '/tmp/' + fname+str(order)+'.dec'
        if order :        
            cmd = 'mismatch -o ' + str(order) + ' ' + fname  + ' ' + out_fname
            os.system(cmd)

            # ??? suffixe  suf = '_'+str(n)
            # ??? Phantom
            out_seqset = SequenceSet.SequenceSet( out_fname, sep=' ', phantom=1)
            return out_seqset
        """
        # Build a String List Argument
        seqstrings=[]
        for seq in seqset_.sites:
            str = ""
            for i in seqset_.sites[seq]:
                str = str + i
            seqstrings.append( str )
                
        StringList = MS4_lib.NLDMismatch( seqstrings, order )
        
        out_seqset = SequenceSet.SequenceSet( SeqSet=seqset_,
                                              StringList=StringList,
                                              phantom=1)
        
        return out_seqset
       

