#! /usr/bin/env python

# -*- coding: iso-8859-1 -*-

##
##  Copyright (C) 2010 E. Corel, G. Didier, C. Devauchelle, G. Grasseau
##
##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or (at
##  your option) any later version.
## 
##  This program is distributed in the hope that it will be useful, but
##  WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
## 
##  You should have received a copy of the GNU General Public License
##  along with this program; if not, write to the Free Software
##  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
## 
##

from optparse import OptionParser, OptionValueError
from Constants import Const
import re, sys, os
# value = integer
# range = [[integer]:]integer | integer:
# list = integer, integer, ...
# Main.py  -d m|e|me -n range -k value -s value -o a|f|d|w  file-name
# TODO (????) : 
#      - simple NLD 
#      - mode verbose


## Parses the range of order option
#  @param option see OptionParser documentation.
#  @param opt_str see OptionParser documentation.
#  @param value see OptionParser documentation.
#  @param parser see OptionParser documentation.

## Parses range options, and returns the range 
#  in the "option.dest" attribute of parser.values 
#  object (i.e. config)
#  Example option -n 2:4 :
#     + option.dest="order"
#     + set "order" atribute of parser.values to
#       the list [2,4] 
#  @param option see OptionParser documentation.
#  @param opt_str see OptionParser documentation.
#  @param value see OptionParser documentation.
#  @param parser see OptionParser documentation.
def checkIntValue(option, opt_str, value, parser,**kwargs):

    # nmin = kwargs[ "min" ]
    # nmax = kwargs[ "max" ]
    nmin = 0
    nmax = 100000
    val = None
    try:
        val = parser.rargs[0]
        # Remove from rlist the current arguments
        del parser.rargs[0]

    except IndexError:
        sys.exit(kwargs['error'])

    # Empty argument value (n value not specified)
    if not val: 
      sys.exit(kwargs['error'])  
    else:
        n = int( val )
        if (( n < nmin) | (n > nmax)):
            raise OptionValueError("Bad value for %s ." % opt_str )
        else:
            setattr(parser.values, option.dest, n)     


def checkFloatValue(option, opt_str, value, parser,**kwargs):

    nmin = kwargs[ "min" ]
    nmax = kwargs[ "max" ]
    val = None
    try:
        val = parser.rargs[0]
        # Remove from rlist the current arguments
        del parser.rargs[0]

    except IndexError:
        pass
        #sys.exit(kwargs['error'])

    # Empty argument value (n value not specified)
    if not val: 
      sys.exit(kwargs['error'])  
    else:
        n = float( val )
        if (( n < nmin) | (n > nmax)):
            raise OptionValueError("Bad value for %s ." % opt_str )
        else:
            setattr(parser.values, option.dest, n)     


## Parses the range of order option
#  @param option see OptionParser documentation.
#  @param opt_str see OptionParser documentation.
#  @param value see OptionParser documentation.
#  @param parser see OptionParser documentation.

## Parses range options, and returns the range 
#  in the "option.dest" attribute of parser.values 
#  object (i.e. config)
#  Example option -n 2:4 :
#     + option.dest="order"
#     + set "order" atribute of parser.values to
#       the list [2,4] 
#  @param option see OptionParser documentation.
#  @param opt_str see OptionParser documentation.
#  @param value see OptionParser documentation.
#  @param parser see OptionParser documentation.
def checkRange(option, opt_str, value, parser):
    nmin = None;
    nmax = None;
    try:
        value = parser.rargs[0]
        # Remove from rlist the current arguments
        del parser.rargs[0]

    except IndexError:
        pass

    # Empty argument value (n value not specified)
    if not value: 
        nmin = Const.MinOrder
        nmax = Const.Infinity
    elif value.find(':') == -1:    
        # No range - only one value set
        try:
            n = int(value)
            nmin = n
            nmax = n
        except ValueError:
            pass
    else:                                      
        # Range case : separator specified
        m,n = value.split(':')
        if m == '':                        
            # No lower value, set to Const.MinOrder
            nmin = Const.MinOrder
        else:                                 
            # Convert lower value m to integer nmin
            try:
                nmin = int(m)
            except ValueError:
                pass
        if n == '':                        
            # No upper value, set to Const.Infinity
            nmax = Const.Infinity
        else:                               
            # Convert upper value n to integer nmax
            try:
                n_var = int(n)
                if nmin <= n_var:
                    nmax = n_var
            except ValueError:
                pass
    # Test final values
    if ( nmin < Const.MinOrder ) :
        nmin = None
    if ( nmax != Const.Infinity ) :
        if ( nmax < Const.MinOrder ) :
            nmax = None

    # order_range is set to tuple (nmin,nmax)
    if (not nmin or not nmax) :
        raise OptionValueError("Bad range value for %s ." % opt_str )
    else :
        setattr(parser.values, option.dest, [nmin,nmax])     


## Parses sub-options, and returns the value 
#  in the "option.dest + kwargs[.]" attribute of 
#  parser.values object (i.e. config)
#  Example option -o w :
#     + option.dest ="outputs"
#     + kwargs["w"] = "Window"
#     + set "outputsWindow" attribute of parser.values
#       to True
#  @param option see OptionParser documentation.
#  @param opt_str see OptionParser documentation.
#  @param value see OptionParser documentation.
#  @param parser see OptionParser documentation.
def checkMode( option, opt_str, value, parser,**kwargs):

    # Get the possible option list and Remove "error" item
    option_list = kwargs.keys() 
    option_list.remove("error")
    
    # Clear default options
    for k in  option_list   :
        setattr(parser.values, option.dest +kwargs[k], False )

    # Change the configuration    
    for k in  range( len( value) )  :
        if value[k] in option_list :
            setattr(parser.values, option.dest +kwargs[value[k]],  True )
        else :
            sys.exit(kwargs['error'])

## Parses sub-options, and returns the value 
#  in the "option.dest + kwargs[.]" attribute of 
#  parser.values object (i.e. config)
#  The modes are exclusive i.e. 2 or more cannot be selected
#  Example option -o w :
#     + option.dest ="outputs"
#     + kwargs["w"] = "Window"
#     + set "outputsWindow" attribute of parser.values
#       to True
#  @param option see OptionParser documentation.
#  @param opt_str see OptionParser documentation.
#  @param value see OptionParser documentation.
#  @param parser see OptionParser documentation.
def checkExclusiveMode( option, opt_str, value, parser,**kwargs):

    # Get the possible option list and Remove "error" item
    option_list = kwargs.keys() 
    option_list.remove("error")
    
    # Clear default options
    for k in  option_list   :
        setattr(parser.values, option.dest +kwargs[k], False )

    # Change the configuration    
    if ( len(value) > 1 ) :
        sys.exit(kwargs['error'])
    else :
        if value in option_list :
            setattr(parser.values, option.dest +kwargs[value],  True )
        else :
            sys.exit(kwargs['error'])

## Parses sub-options, and returns the value 
#  in the "option.dest + kwargs[.]" attribute of 
#  parser.values object (i.e. config)
#  The modes are exclusive i.e. 2 or more cannot be selected
#  Example option -o w :
#     + option.dest ="outputs"
#     + kwargs["w"] = "Window"
#     + set "outputsWindow" attribute of parser.values
#       to True
#  @param option see OptionParser documentation.
#  @param opt_str see OptionParser documentation.
#  @param value see OptionParser documentation.
#  @param parser see OptionParser documentation.
def checkSeqSubSetFile( option, opt_str, value, parser,**kwargs):
    val = None
    name_list = []
    k = 0
 
    if not value: 
      sys.exit(kwargs['error'])  
    else:
        try:
            f = open( value )
            for line in f.xreadlines():
                # Not empty line and doesn't start with '#'
                if( (len( line ) != 0) and ( not line.startswith('#') ) ):
                    # Remove '\n' an append in list
                    name_list.append( line.replace('\n', '') )
                    k = k+1
            f.close()
            setattr(parser.values, "inputSeqSubSetFName", value)  
            setattr(parser.values, option.dest, name_list)  
        except IOError :
            sys.exit(kwargs['error_f'])  
        
def GetConfig():

    usage = "usage: %prog [-d e|m] [-n RANGE] [-k value] \
[-s value] -f seq_filter_file [-o a|f|d|w] fasta_file\n \
        RANGE : [[integer]:]integer | integer:"
    parser = OptionParser(usage=usage)

    # Set the default configuration "config"
    parser.set_defaults(
        
        # Parameters
        modeExact=True, modeMismatch=False, 
        order=
        [2, Const.Infinity], 
        orderStr=[None, None], 
        sMin=[2,2],
        kappa = 1.0,

        # Input files
        inputFName=None,
        inputSeqSubSetFName = '', 
        inputSeqSubSet      = [],
        
        # Output files
        outputsAll=False, outputsDistance=False, outputsFasta=True, 
        outputswindow=False, outputsWindow=False, outputsPTree=False, 
        outputsDName=None, outputsFName=None, 
        outputsWName=None, outputsPName=None,

        # Internal use
        TmpFiles=[]
        )
    
    parser.add_option("-d", "--decodemode", type="str", dest="mode", 
                      action="callback", callback=checkExclusiveMode, 
                      callback_kwargs={  
                         'e' : 'Exact', 		 
                         'm' : 'Mismatch', 
                         'error' : 'Wrong type of decoding (e/m)'}, 
                      help="type of decoding exact or mismatched [e/m]", 
                      metavar="MODE")

    parser.add_option("-n", "--order", dest="order",  action="callback", 
                      callback=checkRange,  
                      help="order range of N-local decoding [nmin:nmax]", 
                      metavar="RANGE")

    parser.add_option("-k", "--kappa", dest="kappa", 
                      action="callback", callback=checkFloatValue,
                      callback_kwargs={  
                          'min' : 0,
                          'max' : Const.Infinity,
                          'error' : 'Kappa must be positive'},
                      help="Kappa criteriom to prune the Partition Tree ", 
                      metavar="VALUE")

    parser.add_option("-f", "--seq_sub_set", type="str",dest="inputSeqSubSet", 
                      action="callback", callback=checkSeqSubSetFile, 
                      callback_kwargs={
                        'error' : 'bad option sequence sub set',
                        'error_f' : 'cannot read file of sequence sub set'
                        }, 
                      help="File of sequence sub set extracted from "
                      + "sequence input file",
                      metavar="FILE")

    parser.add_option("-o", "--outputs", type="str",dest="outputs", 
                      action="callback", callback=checkMode, 
                      callback_kwargs={
                        'a' : 'All',
                        'd' : 'Distance', 
                        'f' : 'Fasta', 
                        'w' : 'window', 
                        'W' : 'Window', 
                        'p' : 'PTree',
                        'error' : 'bad option (d|f|w|W|p)'
                        }, 
                      help="Write the distance matrix (d) \n" +
               "or/and write sequences with anchors in a fasta file (f) " +
               "or/and write only anchors and the neighbouring of them (w)" +
               "or/and write position and the neighbouring of the anchors (W)"+
               "or/and write the ptree in .dot format (p)",
                      metavar="MODE")

    parser.add_option("-s", "--smin", dest="sMin", 
                      action="callback", callback=checkRange,
                      help="smin threshold to filter anchors [min:max] ", 
                      metavar="RANGE")

    config, args = parser.parse_args()	
    
    # Set the config.orderStr used for printing, ...
    for i in [0,1] :
        if ( config.order[i] == Const.Infinity ) :
            config.orderStr[i] = "inf";
        else :
            config.orderStr[i] = str(config.order[i]);
            
    # Put config attributes in dictionnary "Attributes"
    Attributes = vars( config )
   
    #  Test input file and store its name in "config"
    file_name = "input.txt"
    DefaultFileNames = True;
    if( len(args) ) :
        file_name = args[ len(args) - 1] 
        try :
            f = open( file_name ) 
            f.close()
            setattr(config, "inputFName", file_name)
            DefaultFileNames = False;
        except IOError :
            print "Error : can't read %s" % file_name
            exit(1)
    else:
        try :
            f = open( file_name ) 
            f.close()
            setattr(config, "inputFName", file_name)
        except IOError :
            print "Error : can't read %s" % file_name
            exit(1)

    setattr(config, "defaultGenericNames", DefaultFileNames)
    #
    #  Generate the names of ouput files in "config"
    #
    generic_name = os.path.basename(file_name).rsplit('.',1)[0]
    
    # Mode case
    mode_field   = None
    if( config.modeExact )    : mode_field = ".e"
    if( config.modeMismatch ) : mode_field = ".m"

    # Order case
    order_field = ".n"
    order_field = order_field + config.orderStr[0]
    order_field = order_field + "-" + config.orderStr[1]
    
    # Kappa value
    kappa_field = ".k"
    kappa_field = kappa_field + str( config.kappa )

    # Smin, Smax
    sMin_field = ".s"
    sMin_field =  sMin_field +  str(config.sMin)

    # Build ouput file names
    dist_name_list = "Output"
    
    if( DefaultFileNames ):
        generic_name = "output"
        dist_name_list  = generic_name + ".nexus" 
        fasta_name_list = generic_name + ".fsa" 
        wind_name_list  = generic_name + ".wdw" 
        ptree_name_list = generic_name + ".dot" 
    else:
        dist_name_list  = generic_name + mode_field + \
                          order_field + kappa_field + sMin_field + ".nexus" 
        fasta_name_list = generic_name + mode_field + \
                          order_field + kappa_field + sMin_field + ".fsa" 
        wind_name_list  = generic_name + mode_field + \
                          order_field + kappa_field + sMin_field + ".wdw" 
        ptree_name_list  = generic_name + mode_field + \
                           order_field + kappa_field + ".dot" 
    
    print "Configuration"
    print "-------------"
    print "  decoding      [e|m]: "
    print "  + exact            : ",  config.modeExact
    print "  + mismatch         : ",  config.modeMismatch
    print "  order     [min:max]: ",  config.order
    print "  kappa       [value]: ",  config.kappa
    print "  s           [value]: ",  config.sMin
    print "  Input Fasta file   : ",  config.inputFName
    if( config.inputSeqSubSetFName != '' ):
        print "  Seq. filter file   : ",  config.inputSeqSubSetFName
    
    print "  outputs [a|d|f|w|p]: "
    if( config.outputsDistance ) :
        print "  + distances (nexus): ", dist_name_list 
    if( config.outputsFasta ) :
        print "  + anchors (fasta)  : ", fasta_name_list
    if( config.outputswindow | config.outputsWindow ) : 
        print "  + anchors (window) : ", wind_name_list
    if( config.outputsPTree ) : 
        print "  + anchors (window) : ", ptree_name_list

    if( DefaultFileNames ):
        generic_name = "output"
        dist_name_list  = generic_name + ".nexus" 
        fasta_name_list = generic_name + ".fsa" 
        wind_name_list  = generic_name + ".wdw" 
        ptree_name_list = generic_name + ".dot"
    else:
        dist_name_list  = generic_name + mode_field + \
                          order_field + kappa_field + ".s%s" + ".nexus" 
        fasta_name_list = generic_name + mode_field + \
                          order_field + kappa_field + ".s%s" + ".fsa" 
        wind_name_list  = generic_name + mode_field + \
                          order_field + kappa_field + ".s%s" + ".wdw" 
        ptree_name_list  = generic_name + mode_field + \
                           order_field + kappa_field + ".dot"

    setattr(config, "outputsDName", dist_name_list )
    setattr(config, "outputsFName", fasta_name_list )
    setattr(config, "outputsWName", wind_name_list )
    setattr(config, "outputsPName", ptree_name_list )

    return config
 
if __name__ == "__main__":
    GetConfig()
