#! /usr/bin/env python

# -*- coding: iso-8859-1 -*-

##
##  Copyright (C) 2010 E. Corel, G. Didier, C. Devauchelle, G. Grasseau
##
##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or (at
##  your option) any later version.
## 
##  This program is distributed in the hope that it will be useful, but
##  WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
## 
##  You should have received a copy of the GNU General Public License
##  along with this program; if not, write to the Free Software
##  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
## 
##


# Importation de modules

import Tree,SequenceSet
import MS4_lib

import re, sys, os, altgraph,itertools
#import boost.graph as bgl
# import networkx as NX
from altgraph import Graph, GraphAlgo, Dot
from collections import deque
import string,time

##
# SuffixTree
#

# ???
# SuffixTree -> STree ??
class SuffixTree(Tree.Tree):


    # GG Mettre un SequenceSet
    def __init__(self, SeqSet=None, graph=None,option='sites'):
        """ Initialises the SuffixTree structure -- from file fname or graph
        option == 'sites' : leaves are couples(seq,pos) and inner nodes integers
        option == 'positions' : leaves are integers (absolute positions) and inner nodes 'i'+int
        Test on the graph input option
        Convert one into the other with the trueSites method of Decodable """

        if SeqSet or graph:
            self.filename = SeqSet.filename

            if SeqSet:

                # Build a String List Argument
                seqstrings=[]
                for seq in SeqSet.sites:
                    str = ""
                    for i in SeqSet.sites[seq]:
                        str = str + i
                    seqstrings.append( str )

                # Call C routines
                TreeStrings = MS4_lib.getSuffixTree( seqstrings, 0 )

                # Build Tree 
                graph = STreeFromStrings( TreeStrings,
                                          SeqSet=SeqSet, option=option)

            Tree.Tree.__init__(self,graph=graph)
            try:
                self.leafId = self.leafId()
            except:
                self.leafId = ''
        else:
            Tree.Tree.__init__(self)

    def copySTree(self):
        g = self.copyTree()
        try:
            g.filename = self.filename[:]
        except:
            g.filename = None
        return g

    #======== Conversion methods

    def convertNodes(self,mode=None):
        """This method converts nodes of integer type and leaves of site type into 'i'XXX and integers -- and conversely
        if a mode is specified, then the specified conversion is attempted"""

        # GG-begin : strange story ???
        try:
            fname = self.filename[:]
            # GG A new read ??? which takeinto account the shift in the
            # Suffix result
            fsa = SequenceSet.SequenceSet(fname,phantom=1)
            site = fsa.site
            position = fsa.position
        except:
            sys.stderr.write('Cannot convert STree, because no Fasta file is specified\n')
            return self

        # GG Empty constructor
        g = SuffixTree( )
        # GG-end strange story ???

        root = self.getRoot()
        if type(root) == int:             # in this case, nodes are in 'site mode'
            if mode == 'site':              # if we specified 'site' mode, we have nothing to do
                return self
            newNode = lambda x:self.newPosNode(position,x)
        elif type(root) == str:         # in this one, in 'position mode'
            if mode == 'position':           # if we specified 'position' mode, we have nothing to do
                return self
            newNode = lambda x:self.newSiteNode(site,x)
        for e in self.edges:
            g.edges[e] = (newNode(self.edges[e][0]),newNode(self.edges[e][1]),self.edges[e][2])
        for n in self.nodes:
            g.nodes[newNode(n)] = self.nodes[n][:]
        g.root = g.getRoot()
        return g

    def newSiteNode(self,site,node):
        """Transforms a node in 'position mode' into a node in 'site mode'"""
        try:                      # we try to convert node as key of site dictionary -- i. e. when is integer
            return site[node]
        except KeyError:   # exception is raised when node is innerNode 'i'XXX
            return int(node.strip('i'))

    def newPosNode(self,position,node):
        """Transforms a node in 'site mode' into a node in 'position mode'"""
        try:                      # we try to convert node as key of position dictionary -- i.e. when is tuple
            return position[node]
        except KeyError:   # exception is raised when node is innerNode (integer)
            return 'i'+str(node)

    #========= Methods for the suffix links -- assume that the STree is in position mode

    def leavesShift(self,innerNode,shift=1):
        leafList = self.childLeaves(innerNode)
        return map(lambda x:x+shift,leafList)

    def suffixLink(self,innerNode):
        """Returns the node pointed at by the suffix link from innerNode"""
        leafList = self.childLeaves(innerNode)
        try:
            return self.LCA(map(lambda x:x+1,leafList))
        except:
            return self.root

    def suffixTriplet(self,innerNode):
        leafList = self.childLeaves(innerNode)
        shiftedLeaves = map(lambda x:x+1,leafList)
        try:
            sLink = self.LCA(shiftedLeaves)
            newLeaves = self.childLeaves(sLink)
            signif = min(1,len(newLeaves)-len(shiftedLeaves))
        except:
            sLink = self.root
            newLeaves = self.getLeaves()
            signif = 1
        #return set(shiftedLeaves), sLink, set(newLeaves),signif
        return sLink,signif

    def sLinks(self,out=''):
        """ Computes the tree of links for 1-shifts, and puts the flag to 0 if node is irrelevant --
        Beware, I think it does not handle correctly the leaves of suffix link tree """
        g = self.convertNodes(mode='position')
        root = g.root
        t0 = time.clock()
        s = g.copySTree()
        tree = Tree.Tree()
        tree.root = g.getRoot()
        flag = dict([(node,1) for node in g.innerdfs(root)])        # here we store relevance of node (wrt suffix links)
        flag[root] = 1
        sys.stderr.write('temps intermediaire '+str(time.clock()-t0)+'\n')
        for n in g.innerdfs():
            #            if flag[n] == 1:                              # if you submit n to this test some irrelevant nodes receive no suffix link and are therefore considered significant
            #dn,m,dm,fl = g.suffixTriplet(n)
            m,fl = g.suffixTriplet(n)
            flag[m] = fl
            newEdge = max(s.edges.keys())+1
            s.edges[newEdge] = (n,m,0)
            s.nodes[n][1].append(newEdge)
            s.nodes[m][0].append(newEdge)
            s.next_edge += 1
            #if not fl:
            #    for nn in g.innerdfs(m):
            #        flag[nn] = 0
            tree.add_edge(m,n)          #V0
            #else:
            #    pass
        sys.stderr.write('Computation took '+str(time.clock()-t0)+'\n')
        # g.viewTree()
        # tree.viewTree()
        if out == 'tree':
            return flag,tree
        else:
            return flag

    def signifDFS(self):
        """ A DFS implementation of non-significant nodes """
        g = self.convertNodes(mode='position')
        t0 = time.clock()
        s = g.copySTree()
        tree = Tree.Tree()
        flag = {}
        sys.stderr.write('temps intermediaire '+str(time.clock()-t0)+'\n')
        father = lambda x:g.inc_nbrs(x)[0]
        stack = map(father,g.getLeaves())
        while stack:
            stack.sort(key=g.node_data)
            stack.reverse()
            newStack = []
            for node in stack:
                if not flag.has_key(node):
                    flag[node] = 1
                print 'node',node,flag[node]
                m,fl = g.suffixTriplet(node)
                flag[m] = fl
                print 'sLink',m,fl
                newEdge = max(s.edges.keys())+1
                s.edges[newEdge] = (node,m,0)
                s.nodes[node][1].append(newEdge)
                s.nodes[m][0].append(newEdge)
                s.next_edge += 1
                try:
                    next_node = father(node)
                    try:
                        i = flag[next_node]
                    except:
                        i = 0
                    if i == 0:
                        newStack.append(next_node)
                    else:
                        for n in g.iterAncestors(node):
                            flag[n] = 1
                except:
                    pass
            stack = newStack
        sys.stderr.write('Computation took '+str(time.clock()-t0)+'\n')
        return flag

    def signifInnerNodes(self):
        """ A depth layer implementation of significant nodes -- more or less ok """
        g = self.convertNodes(mode='position')
        t0 = time.clock()
        s = g.copySTree()
        tree = Tree.Tree()
        depth = g.nodeCounts()
        depths = depth.keys()
        depths.sort()
        depths.reverse()
        print depths
        father = lambda x:g.inc_nbrs(x)[0]
        flag = {}
        sys.stderr.write('temps intermediaire '+str(time.clock()-t0)+'\n')
        for d in depths:
            stack = depth[d]
            for node in stack:
                if not flag.has_key(node):
                    flag[node] = 1
                #print 'node',node,flag[node]
                m,fl = g.suffixTriplet(node)
                flag[m] = fl
                #print 'sLink',m,fl
                newEdge = max(s.edges.keys())+1
                s.edges[newEdge] = (node,m,0)
                s.nodes[node][1].append(newEdge)
                s.nodes[m][0].append(newEdge)
                s.next_edge += 1
                try:
                    next_node = father(node)
                    try:
                        i = flag[next_node]
                    except:
                        i = 0
                    if i != 0:
                        for n in g.iterAncestors(node):
                            flag[n] = 1
                except:
                    pass
        sys.stderr.write('Computation took '+str(time.clock()-t0)+'\n')
        return flag

    def DFSparse(self,f,start=None):
        """Iterator on all nodes of tree pruned by condition f -- which is a (boolean) function defined on the nodes.
        Explicitly : starting from 0, outputs the nodes in DFS while condition is True, where it yields (node,1), and
        at the first False value, yields (node,0)
        Note : if f is not monotonic, nothing tells that this returns all nodes satisfying f == True"""
        if not start:
            start = self.root
        # DFS starts here
        visited, stack = set([start]), deque([start])
        get_edges = self.out_edges
        while stack:
            curr_node = stack.pop()
            try:
                u = f(curr_node)
            except:
                u = 1
            if u != 0:
                if curr_node != start:
                    yield (curr_node,1)
                for edge in get_edges(curr_node):
                    tail = self.tail(edge)
                    if tail not in visited:
                       visited.add(tail)
                       #if self.out_nbrs(tail):    # comment if you want to keep the leaves that are marked
                       stack.append(tail)
            else:
                yield (curr_node,0)
                if curr_node not in visited:
                    visited.add(curr_node)

    def DFSInnerParse(self,f,start=None):
        """Iterator on all nodes of tree pruned by condition f -- which is a (boolean) function defined on the nodes.
        Explicitly : starting from 0, outputs the nodes in DFS while condition is True, where it yields (node,1), and
        at the first False value, yields (node,0)
        Note : if f is not monotonic, nothing tells that this returns all nodes satisfying f == True"""
        if not start:
            start = self.root
        # DFS starts here
        visited, stack = set([start]), deque([start])
        get_edges = self.out_edges
        while stack:
            curr_node = stack.pop()
            try:
                u = f(curr_node)
            except:
                u = 1
            if u != 0:
                if curr_node != start:
                    yield (curr_node,1)
                for edge in get_edges(curr_node):
                    tail = self.tail(edge)
                    if tail not in visited:
                       visited.add(tail)
                       if self.out_nbrs(tail):    # comment if you want to keep the leaves that are marked
                           stack.append(tail)
            else:
                yield (curr_node,0)
                if curr_node not in visited:
                    visited.add(curr_node) 

    def signifTree(self,out=None):
        """returns set of significant part of suffix tree"""
        flag,tree = self.sLinks(out='tree')
        tree.saveTree(fileout='sLinks')
        s = SuffixTree()
        for n in self.iterDFS():   # our definition, excludes root, but also parses leaves
            try:
                par = self.inc_nbrs(n)[0]  # if not root
                try:
                    a = flag['i'+str(n)]           # has flag (ie inner node), computed in 'position mode'
                    d = self.node_data(par)
                    if a == 0:     # if non relevant 
                        for leaf in self.childLeaves(n):   # add leaves to parent
                            s.add_edge(par,leaf)
                    else:            # if relevant
                        s.add_edge(par,n)  # add normal edge
                except KeyError:    # if has not flag (ie leaf)
                    s.add_edge(par,n)   # add edge
            except IndexError:        # if root do not do anything
                pass
        # s.viewTree()
        return s

    def signifNodes(self):
        """ Iterator on significant nodes """
        flag = self.signifInnerNodes()
        signif = lambda x:flag['i'+str(x)]
        return self.DFSInnerParse(signif)

    def signifTree1(self,out=None):
        """ Returns set of significant part of suffix tree -- the good one """
        flag = self.sLinks()
        s = SuffixTree()
        signif = lambda x:flag['i'+str(x)]           # has flag (ie inner node), computed in 'position mode']
        for n,sig in self.DFSparse(signif):
            try:
                par = self.inc_nbrs(n)[0]  # if not root
                try:
                    d = self.node_data(par)
                    if sig == 0:     # if non relevant 
                        for leaf in self.childLeaves(n):   # add leaves to parent
                            s.add_edge(par,leaf)
                    else:            # if relevant
                        s.add_edge(par,n)  # add normal edge
                except KeyError:    # if has not flag (ie leaf)
                    s.add_edge(par,n)   # add edge
            except IndexError:        # if root do not do anything
                pass
        if out:
            s.viewTree()
        s.saveTree(fileout='STree-sign')
        return s

    def signifValues(self,out=None,opt=''):
        """returns set of significant values of suffix tree"""
        s = set([])
        for n,sig in self.signifNodes():
            if sig == 1:
                s.add(self.node_data(n))
        l = list(s)
        l.sort(cmp=lambda x,y:x-y)
        if out:
            string = ''
            f = open(out,'w')
            for n in l:
                string += str(n)+' '
            f.write(string+'\n')
            f.close()
        return l

#========== "New" version of decode

    def initEq(self,n=1,opt=0):
        """Define initial equivalence relation R_0^N. Returns dictionary
        c : leaf -> current class index (initially, ancestor node id)"""
        c = {}
        f = lambda x:(self.node_data(x) >= n)
        for v in self.DFSpruneBorder(f):
            for l in  self.childLeaves(v):
                c[l] = v
        return c

#========== sTree
# TODO : obsolete
def sTree(treefile,name='Pos',fasta=None,option='sites'):
    """Computes the compacted suffix tree with depths from GD's output from printSuffixTree which is by default 'SuffixTree.out'
    Leaf identifier is integer corrsponding to absolute position in SuffixTree.out, and node identifier is 'i'XXX
    if fasta == fastaName, then one constructs them as sites = (seq,pos) included the "imaginary" positions (seq,seqLength)"""
    fsa = SequenceSet.SequenceSet(fasta,phantom=1)
    if option == 'sites':                  # if 'sites', then leaves are sites, and nodes integers
        leaf = lambda x:fsa.site[x]
        node = lambda x:int(x)
    elif option == 'positions':      # if 'positions', then leaves are integers, and nodes 'i'+str(n)
        node = lambda x:'i'+str(x)
        leaf = lambda x:int(x)
    f = open(treefile,'U')
    f.readline()
    g = Graph.Graph()
    k = 0
    root = node(k)
    g.add_node(root,0)            # root = 0
    path = [(root,0)]                 # in path, we store the list of (node,depth) that correspond to the compacted STree structure in 'SuffixTree.out'
    for l in f.xreadlines():  # we 'strip' a 'SuffixTree.out' line in its elements : nb of '---' and description of edge label ['begin', 'end'] into 'begin' and 'end'
        l = l.strip('\n')
        prof,suff = l.split('+')
        d = len(prof)                 # nb of '-' in the read line
        sf = suff.strip('[]').split(', ')
        b,e = int(sf[0]),sf[1]   # begin and end of edge label
        path = path[:d]           # restrict the path to the current node
        cv,dd = path[-1]          # extract coordinates (int, depth) of last node
        if e != 'infty':                # if edge label is not suffix
            k += 1                        # update label for inner node
            vertex = node(k)
            l = int(e)-b+1           # compute length of edge label
            w = dd+l                   # Integer which stores the real depth of node = previous depth + edge label length
            path.append((vertex,w))# add new node descriptor in path
            g.add_node(vertex,w)     # add node with data=depth...
            g.add_edge(cv,vertex,l)  # ...and edge with data=label length in STree
        else:                                # if label is suffix
            i = b-dd                     # compute position of suffix
            leafName = leaf(i)
            g.add_node(leafName)         # finish STree by adding leaves and edges to leaves
            g.add_edge(cv,leafName,'')
    f.close()
    return g

# Build SuffixTree from sequence strings
def STreeFromStrings(TreeStr, SeqSet, option='sites'):
    """Computes the compacted suffix tree with depths from GD's output from printSuffixTree which is by default 'SuffixTree.out'
    Leaf identifier is integer corrsponding to absolute position in SuffixTree.out, and node identifier is 'i'XXX
    if fasta == fastaName, then one constructs them as sites = (seq,pos) included the "imaginary" positions (seq,seqLength)"""

    # fsa = SequenceSet.SequenceSet(fasta,phantom=1)
    # The C code of NLD use position >= 1
    # So need a shift set in the Mapping
    var =  [s for s in SeqSet.siteIterator(phantom=1)]
    Mapping = dict(zip(range(len(var)),var))
    
    if option == 'sites':
        # if 'sites', then leaves are sites, and nodes integers
        leaf = lambda x:Mapping[x]
        node = lambda x:int(x)
    elif option == 'positions':
        # if 'positions', then leaves are integers, and nodes 'i'+str(n)
        node = lambda x:'i'+str(x)
        leaf = lambda x:int(x)
    g = Graph.Graph()
    k = 0
    root = node(k)
    g.add_node(root,0)
    # in path, we store the list of (node,depth) that correspond to
    # the compacted STree structure in 'SuffixTree.out'
    path = [(root,0)]
    # Skip the first line
    for l in TreeStr[1:] :
    # for l in TreeStr :
        # we 'strip' a line in its elements : nb of '---'
        # and description of edge label ['begin', 'end']
        # into 'begin' and 'end'
        l = l.strip('\n')
        prof,suff = l.split('+')
        d = len(prof)                 # nb of '-' in the read line
        sf = suff.strip('[]').split(', ')
        b,e = int(sf[0]),sf[1]   # begin and end of edge label
        path = path[:d]           # restrict the path to the current node
        # print "path :", path
        cv,dd = path[-1]          # extract coordinates (int, depth) of last node
        if e != 'infty':                # if edge label is not suffix
            k += 1                        # update label for inner node
            vertex = node(k)
            l = int(e)-b+1           # compute length of edge label
            w = dd+l                   # Integer which stores the real depth of node = previous depth + edge label length
            path.append((vertex,w))# add new node descriptor in path
            g.add_node(vertex,w)     # add node with data=depth...
            g.add_edge(cv,vertex,l)  # ...and edge with data=label length in STree
            # print "new node = ", vertex, ", ", cv, "->", vertex
        else:                                # if label is suffix
            i = b-dd                       # compute position of suffix
            leafName = leaf(i)
            g.add_node(leafName)         # finish STree by adding leaves and edges to leaves
            g.add_edge(cv,leafName,'')
            # print "new leaf = ", leafName, ", ", cv, "->", leafName
    return g

