#! /usr/bin/env python

# -*- coding: iso-8859-1 -*-

##
##  Copyright (C) 2010 E. Corel, G. Didier, C. Devauchelle, G. Grasseau
##
##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or (at
##  your option) any later version.
## 
##  This program is distributed in the hope that it will be useful, but
##  WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
## 
##  You should have received a copy of the GNU General Public License
##  along with this program; if not, write to the Free Software
##  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
## 
##

# Importation de modules

# import tG,mySyntax
import mySyntax
import re, sys, os, altgraph,itertools
from altgraph import Graph, GraphAlgo, Dot
from collections import deque
import string

# ??? Subtitutions
# FastaObject -> SequenceSet
# .readFastaSites -> .readFastaFile
# .sequences -> .SeqNames
# .removeSeq -> writeRemFastaFile
# .recoverFasta -> writeFastaFile
# .letterLocation -> getECLocation
class SequenceSet:

    def __init__(self, fname=None, SeqSet=None, StringList=None,
                 sep='', phantom=None):
        if fname:
            # Types 
            #   SeqNames : list of sequence names,
            #   sites    : dict( Id sequence, list of letters  )
            #   code     : dict( sequence name, Id sequence)
            #            ??? SeqNameToId
            #   site     : dict ( Id letter relative to the file (sequenceset), 
            #              ( Id sequence, Id letter relative to the sequence )
            #            ??? IdToRelativeId
            #   site     : dict ( Id letter relative to the file (sequenceset), 
            #   position : dict( ( Id sequence, Id letter relative to 
            #                      the sequence ), 
            #                    Id letter relative to the file/sequenceset ) 
            #            ??? RelativeIdToId
            # ??? verifier que l'on enleve bien les separateurs

            self.SeqNames, self.sites = self.readFastaFile(fname,sep)
            self.code = dict([(seq,self.SeqNames.index(seq)) for seq in self.SeqNames])
            self.site = self.positionsToSites(opt=phantom)
            self.position = self.sitesToPositions(opt=phantom)
        elif SeqSet and StringList :
            self.SeqNames = SeqSet.SeqNames
            # Transform to dictonnary
            self.sites    = {}
            k = 0
            for i in StringList:
                # remove last character (a "space character")
                self.sites[k] = string.split( i[:-1]," ")
                k = k+1
            self.code     = dict([(seq,self.SeqNames.index(seq))
                                  for seq in self.SeqNames])
            self.position = self.sitesToPositions(opt=phantom)
        else:
            self.SeqNames = []
            self.sites = {}
            self.code = {}
            self.site = {}
            self.position = {}
        self.filename = fname

    #====== Attribute readers

    def getSeqNames(self):
        return self.SeqNames

    def getSites(self):
        return self.sites

    #======= Syntaxic methods

    def copy(self):
        """Copy method"""
        g = SequenceSet()
        g.SeqNames = self.SeqNames[:]
        g.sites = mySyntax.copyDict(self.sites)
        g.code = mySyntax.copyDict(self.code)
        g.site = mySyntax.copyDict(self.site)
        g.position = mySyntax.copyDict(self.position)
        g.filename = self.filename[:]
        return g

#   Extract a subset of SequenceSet.
#   Arguments :
#   - selected_seq : list of sequence names
#   - verbose : ...
    def extractSubSet(self, selected_seq = [], verbose=True):
        
        g = SequenceSet()
        SelSeqID = []
        count = 0
        if ( selected_seq != [] ):
            # Build seq index list
            for i in selected_seq:
                try:
                    SelSeqID.append( self.SeqNames.index( i ) )
                    count = count + 1
                except ValueError :
                    SelSeqID.append(-1 )
                
            k = 0        
            for i in range( len( SelSeqID ) ) :
                if ( SelSeqID[i] != -1 ) :
                    g.SeqNames.append( selected_seq[i] )
                    g.sites[k] = self.sites[ SelSeqID[i] ]
                    k = k + 1

            # Initialize the other object members
            g.code     = dict([(seq, g.SeqNames.index(seq))
                               for seq in g.SeqNames])
            g.site     = g.positionsToSites()
            g.position = g.sitesToPositions()
            g.filename = self.filename + ".sub"
        else :
            g = self.copySequenceSet( )
            
        if ( verbose ) :
        
            sys.stdout.write( "SequenceSet: " + str(count) +
                              " sequence(s) extracted from " +
                              str( len( self.SeqNames ) ) +
                              " sequence(s)" + "\n")

        return g

    
    # ??? Doit etre fusionner avec writeRemFastaFile
    def writeRemFastaFile(self,seq,extension='.rm',separator=''):
        """Removes the sequences seq, and writes the Fasta file"""
        g = self.copySequenceSet() 
        if not isinstance(seq,list):
            seq = [seq]
        for s in seq:
            del g.sites[g.code[s]]
            g.SeqNames.remove(s)
        g.writeFastaFile('',extension,separator)
        

    def writeFastaFile(self,name=None,ext=None,sep=' ',lineLength=80):
        """
        Reconstructs a fasta file from the SequenceSet -- dir = directory, ext = extension, to recognize it from the input, sep=to separate characters (GD classes) """

        if not isinstance(name, str): 
            sys.exit("writeFastaFile: No file name")

        if isinstance(ext, str): 
            name = name + ext

        f = open( str(name),'w')
        for seq in self.SeqNames:
            line1 = '>%s\n'% (seq)   
            f.write(line1)                    
            lineList = self.sites[self.code[seq]]  
            i = 0                                                         
            newline = ''
            for l in lineList:  
                if l:
                    newline += str(l)+sep                               
                    i += 1
                    try:
                        if i % int(lineLength) == 0:
                            newline += '\n' 
                            f.write(newline) 
                            newline = ''
                    except ZeroDivisionError:
                        pass
            f.write(newline+'\n')    
        f.close()
                
        """ GGG ??? - constructor """
     #=========  Constructor methods
                            
    def siteIterator(self,phantom=None):
        """Returns an iterable on (seq,pos) with seq repeated as many times as sites in the sequence"""
        sit = {}
        ss = []
        dec = 0
        if phantom: # phantom = sthing means that we also want to include the phantom character (seq,length of seq)
            dec = 1
        s = len(self.SeqNames)
        for i in range(s):
            sit[i] = len(self.sites[i]) + dec
            a = [i for t in range(sit[i])]
            ss.extend(zip(a,range(sit[i])))
        return iter(ss)


    def positionsToSites(self,opt=None):
        """Constructs the dictionary pos -> (seq,position) for self  -- opt != None means that we include phantom site"""
        sites = [s for s in self.siteIterator(opt)]
        site = dict(zip(range(len(sites)),sites))
        position = dict(zip(sites,range(len(sites))))
        return site

    def sitesToPositions(self,opt=None):
        """Constructs the dictionary (seq,position) -> pos for self -- opt != None means that we include phantom site"""
        sites = [s for s in self.siteIterator(opt)]
        position = dict(zip(sites,range(len(sites))))
        return position

    #======= Decoding methods

    def badIndices(self):
        """List of phantom sites -- (sequence index, length of sequence)"""
        return [(a,len(self.sites[a])) for a in self.sites.keys()]

    def nextSite(self,site,shift=1):
        """For a given site (seq,pos), return the shifted site+shift -- if error returns None"""
        try:
            return self.site[self.position[site]+shift]
        except KeyError:
            return None
        
    def nextSites(self,siteList,shift=1):
        """List version of nextSite"""
        return map(lambda x:self.nextSite(x,shift),siteList)

    def getSite(self,position):
        """This method will recognise when the jump went over the absolute position range"""
        try:
            return self.site[position]
        except KeyError:
            return None

    #====####### Anchor methods

    def getECLocation(self):
        """ GGG get the coordinates of the EC 
        (seq_id[0..n], pos[0..length(sed_id) in the sequenceset """
        """ Forms the dictionary letter -> list of sites (seq,position) bearing letter"""
        classes = {}
        for s,l in self.sites.iteritems():
            for i in range(len(l)):
                try:
                    classes[l[i]].append((s,i))
                except KeyError:
                    classes[l[i]] = [(s,i)]
        return classes

    ## Return the neibouring (string) of n-width at the location "site". 
    #  @param self the object pointer.
    #  @param site the location in the sequence.
    def getNeighbouring(self, site, n, sep=''):

        seq,pos = site
        a,b = max(0,pos-n+1),min(len(self.sites[seq]),pos+n)
        a1,b1 = pos-n+1,pos+n
        letterList = []
        for i in range(a-a1):
            letterList.append('-')
        for i in range(b)[a:]:
            letter = str(self.sites[seq][i])
            letterList.append(letter)
        for i in range(b1-b):
            letterList.append('-')
        letterList[n-1] = sep+letterList[n-1]+sep
        line = mySyntax.ListToString(letterList)
        return line

#======### Definitions essentielles pour le traitement de fichiers FASTA
    # ??? Position douteuse : devrait lire et initialiser le SequenceSet
    def readFastaFile( self, name, sep=' '):
        """This the main function of SequenceSet constructor!!
        Reads a FASTA or enriched FASTA file (separator option) 
        Returns
        1. the list of sequence names
        2. the dictionary index of name -> list of sites of sequence"""	
        sites = {}
        f     = open(name,'U')
        s     = []
        seqs  = []
        for line in f.xreadlines():
            line = line.strip().strip('\n')
            if line.startswith('>'):
                if s:
                    sites[sq] = s[:]
                sq = line.strip('>')
                s = []
                seqs.append(sq)
            elif line:
                if sep == '':
                    s.extend(list(line))
                else:
                    s.extend(line.split(sep))
        sites[sq] = s[:]
        f.close()
        newsites = {}
        for a in sites.keys():
            newsites[seqs.index(a)] = sites[a]
        return seqs,newsites

    ## Return all letters which occur in the 
    #  site list.
    #  @param self the object pointer.
    #  @param site_list list of sites.
    def getLettersOfSites( self, site_list ):

        letterList = []
        for site in site_list:
            try:
                letterList.append(self.sites[site[0]][site[1]])
            except IndexError:
                letterList.append('-')

        letters = list(set(letterList))
        letters.sort(key=str.lower)
        return reduce(lambda x,y:str(x)+str(y),letters)


    def isConform( self, SSet_, strict=True):

        isConform = True;

        # Dictionnary containing mapping between 
        # reference patterns (in self) and those
        # in SSet_
        # Example :  Pattern = { 'G0_5': 'I368_5', 'A0_3': 'I332_3', ... }
        # If 'strict' is set to True verify that terminal characters 
        # match exactly

        Pattern = {};

        # Catching an IndexError exception means that
        # Sequences are differents
        try:
            for i in self.sites.keys():
                for j in range( len( self.sites[i] )):
                    word = self.sites[i][j];
                    if ( word in Pattern.keys() ):
                        # Test matching patterns
                        if( Pattern[ word ] != SSet_.sites[i][j] ):
                            isConform = False; 
                            sys.stderr.write(
                                "SequenceSet.isConform (seq_id, pos): " \
                                    + str(i) + ' ' + str(j) + "\n")
                            sys.stderr.write(
                                "SequenceSet.isConform (word, word): " \
                                    + word +  ' ' + SSet_.sites[i][j] + "\n")
                            
                    else:
                        # New pattern: store in dictionnary pattern
                        Pattern[ word ] = SSet_.sites[i][j];
                            
                  # ???  if( not isConform ):
                  #      break;
                # if( not isConform ):
                #    break;
        except IndexError:
            isConform = False;
            sys.stderr.write(
                "SequenceSet.isConform : Bad sequence size" \
                    + "\n")
            return( isConform ) 

        # Test terminal code (protein alphabet or ADN alphabet)
        for i in Pattern.keys():
            if ( (len(i) == 1) or ( len(Pattern[i]) == 1 )):
                # Terminal case
                if( i != Pattern[i]):
                    sys.stderr.write(
                     "SequenceSet.isConform : Bad terminal character " \
                         + i +  ' ' +  Pattern[i]+ "\n")
                    if( strict ):
                        isConform = False 

        
        return( isConform )

 #??? SSet1 = SequenceSet.SequenceSet( "maude3.e.n2-inf.s2.fsa" ,  sep=" ")
    # ??? sites

# ??? Inv 8/07/09
def mapLetter(letterList):
    """Returns the compound letter grouped by MM decoding"""
    letters = list(set(letterList))
    letters.sort(key=str.lower)
    return reduce(lambda x,y:str(x)+str(y),letters)
