#! /usr/bin/env python

# -*- coding: iso-8859-1 -*-

##
##  Copyright (C) 2010 E. Corel, G. Didier, C. Devauchelle, G. Grasseau
##
##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or (at
##  your option) any later version.
## 
##  This program is distributed in the hope that it will be useful, but
##  WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
## 
##  You should have received a copy of the GNU General Public License
##  along with this program; if not, write to the Free Software
##  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
## 
##

import sys,getopt,os,re

from altgraph import Graph

# la commande magique: for f in $(ls *.llr); do gawk '{if ($2 > 2.0); print $0}' $f > $f.filter2.0; done

def myCompare(x,y):
    if x < y:
        return -1
    elif x > y:
        return 1
    else:
        return 0

def lexicoTuple(x,y):
    """Defines the lexicographic ordering of couples (seq,pos)"""
    try:
        x0,x1 = x
        y0,y1 = y
    except (ValueError,TypeError):
        return myCompare(x,y)
    a = lexicoTuple(x0,y0)
    if a:
        return a
    else:
        return lexicoTuple(x1,y1)

def tupleToDict(siteList):
    """For a lexicoTuple ordered list of sites (tuples), convert it into the dictionary seq -> ordered list of positions in seq"""
    seqSite = {}
    for site in siteList:
        s,p = site
        try:
            seqSite[s].append(p)
        except KeyError:
            seqSite[s] = [p]
    return seqSite

def extractTuple(a,sep='()'):
    return map(lambda x:int(x),a.strip(sep).split(', '))

def extractTuple0(a,sep='()'):
    return a.strip(sep).split(', ')

def tuplify(x):
    """A syntactic routine to tackle a format problem"""
    y = extractTuple(x)
    return (y[0],y[1])
    
def matchTuple(a):
    p = re.compile('\(\d+, \d+\)')
    return map(extractTuple,p.findall(a))

def myZip(l1,l2):
    l = []
    for i in range(len(l1)):
        try:
            ll = l1[i]
            ll.append(l2[i])
        except AttributeError:
            ll = [l1[i]]
            ll.append(l2[i])
        l.append(ll)
    return l

def removeAll(item='', item_list=[]):
    """ This function removes all occurrences of 'item' in 'item_list' """
    while True:
        try:
            item_list.remove(item)
        except ValueError:
            break
    return item_list

def invertDico(dico):
    """
    Builds the reverse dictionary value -> (list of) keys -- not tested for velocity"""
    ocid = {}
    for i,v in dico.iteritems():
        try:
            ocid[v].append(i)
        except IndexError:
            ocid[v] = [i]
    return ocid

def invertDico1(dico):
    """ builds the reverse dictionary value -> (list of) keys """
    lst1 = [(v,i) for i,v in dico.iteritems()]
    print lst1
    gr = Graph.Graph()
    for e in lst1:
        gr.add_edge(*e)
    lst2 = [(e,gr.out_nbrs(e)) for e in gr.nodes if gr.out_nbrs(e)]
    ocid = dict(lst2)
    return ocid

def subDict(dico,condition):
    subDico = {}
    for key,value in dico.iteritems():
        if condition(key):
            subDico[key] = value
    return subDico

def copyDict(dico):
    """Performs a deep copy of dico"""
    g = {}
    for i,v in dico.iteritems():
        try:
            g[i] = v[:]
        except TypeError:
            g[i] = v
    return g

def partitionUnion(setList,newSet):
    """Computes the partition obtained by adding newSet (or list of newSets) to the setList"""
    #initial test
    if not setList and newSet:
        return [newSet]
    # from now on, we have a non-empty setList
    att = {}                   # this is the element -> index dictionary
    k = 0                       # initialise the first index to 0
    # start by attributing an index to each set in setList (ie the same to all of its elements)
    for s in setList:    
        for v in s:
            att[v] = k
        k += 1
    # the dictionary att is isomorphic to setList
    tta = tG.invertDico1(att)        # tta is the reverse dictionary index -> list of elements in set labelled by 'index'
    # here we compute the list of labels such that set indexed by label has a non-empty intersection with newSet
    labels = set([])
    for i in newSet:
        try:
            labels.add(att[i])
        except KeyError:
            pass
    # if no label is found, then the newSet is disjoint from sets in setList, and will be simply added
    newClass = newSet
    # if an intersection exists, we construct the reunion of sets with indices in labels
    for l in labels:
        nx = set(tta[l])
        newClass.update(nx)
        del tta[l]                 # and destroy the old labels
    # we relabel all elements in newClass with the new label k
    for j in newClass:
        att[j] = k
    # and construct the new description of sets tta
    tta = tG.invertDico1(att)
    newSets = []
    # we now use tta to generate the new setList
    for m in tta.keys():
        s = set(tta[m])
        newSets.append(s)
    return newSets
               
def partitionUnionList(setList,newSetList):
    """List of newSets version of partitionUnion"""
    if not isinstance(newSetList,list):
        newSetList = [newSetList]
    for newSet in newSetList:
        setList1 = partitionUnion(setList,newSet)
        setList = setList1
    return setList

def printDict(dico):
    for i,v in dico.iteritems():
        v.sort(lexicoTuple)
        print i,v

def ListToString(sList,sep=''):
    """Convert list into string of concatenated elements"""
    line = ''
    for s in sList:
        line += sep+str(s)
    return line

#======== Matrix functions

def initMatrix(n):
        mm = [[0]*n for i in range(n)]
        return mm

def printMatrix(a,format='(%3d,%3d)'):
    """Prints a matrix, implemented as a list of lines"""
    l,c = len(a),len(a[0])
    # Formatage de la matrice
    form = '"'+format
    for i in range(c-1):
        form += ' '+format
    form += '"'
    for j in range(l):
        line = ()
        for i in range(c):
            line += (a[j][i],)
        print form % line    

def writeNexus(filename,distances,names,format='%.3f ',comment='',ext='.nexus'):
    """Fonction qui ecrit une matrice au format nexus (cf doc SplitsTree) -- For any complaints, cf. Claudine"""
    # ouverture fichier 
    fname = filename+ext               # rajout extension
    f = open(fname,'w')
    nm = {}
    # ecriture resultat
    
    # header nexus
    nbseq = len(distances)
    f.write('#NEXUS\n\n')
    titre = '[Distance matrix calculated by '+comment+']\n\n'
    f.write(titre)                                  # titre
    f.write('BEGIN taxa;\n')                # labels
    dimension = '\tDIMENSIONS ntax = %d;\n' % (nbseq) # - elt[0]
    f.write(dimension)
    f.write('TAXLABELS\n')
    for i in xrange(0,nbseq):
        nm[i] = names[i].replace(' ','_').replace('/','-')
        f.write('\t'+nm[i]+'\n')
    f.write(';\n')
    f.write('END;\n\n')
    f.write('BEGIN distances;\n')   # header distances
    f.write(dimension)
    f.write('\tFORMAT\n')
    f.write('\t\ttriangle=BOTH\n')
    f.write('\t\tdiagonal\n')
    f.write('\t\tlabels\n')
    f.write(';\n')
    f.write('MATRIX\n')
    
    # matrice de distances
    for i in xrange(0,nbseq):
        ligne = nm[i]+'\t'
        for d in distances[i]:  # parcours une ligne de la matrice (sauf elt0)
            w = format % (d)               # formatage du reel
            ligne = ligne + w       # rajout a la ligne en cours
            ligne = ligne+'\n'
        f.write(ligne)
    f.write(';\n')
    f.write('END;\n')
    f.close()

def msf2fasta(filein,fileout):
    f = open(filein,'U')
    names = []
    fasta = {}
    for line in f.xreadlines():
        l = line.split(' ')
        if l[0] in names:
            try:
                fasta[l[0]].extend(l[1:])
            except:
                fasta[l[0]] = l[1:]
        if line.startswith(' Name:'):
            l = line.split(' ')
            names.append(l[2])
    #print names
    for i in names:
        string = ''
        for el in fasta[i]:
            for letter in el:
                if re.match('\w',letter):
                    string += letter.upper()
        if string:
            fasta[i] = string
        else:
            fasta[i] = string
            msg = 'Attention, sequence %s in %s was removed because it is empty\n' %(i,filein)
            sys.stderr.write(msg)
    #print fasta
    f.close()
    g = open(fileout,'w')
    for i in names:
        if fasta[i]:
            print >> g,'>'+i
            print >> g,fasta[i]
    g.close()
    sys.stderr.write('transformed '+filein+'...\n'+'...into '+fileout+'\n')

def main():
    return

#
# Programme principal
#
#

if __name__  ==  '__main__':
        
    #========
    #
    #                Lecture des arguments
    #
    #========
    
    m = len(sys.argv)                            # lecture des arguments passes en ligne de commande (nombre d'iceux)
    if m >= 1:                                      # mode script : les arguments ont ete passes en ligne de commande
        main(*sys.argv[1:])
    else:                                                  # nombre incorrect d'arguments : fin du script
        sys.exit('Usage : mySyntax.py')


