#! /usr/bin/env python

# -*- coding: iso-8859-1 -*-

##
##  Copyright (C) 2010 E. Corel, G. Didier, C. Devauchelle, G. Grasseau
##
##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or (at
##  your option) any later version.
## 
##  This program is distributed in the hope that it will be useful, but
##  WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
## 
##  You should have received a copy of the GNU General Public License
##  along with this program; if not, write to the Free Software
##  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
## 
##

from ParseArgs import GetConfig
import SequenceSet
import BuildEquivClassesLayers, EquivClassesLayers
import BuildAnchorSet, AnchorSet, AnchorOperator, Matrix
import os, sys 
# import pycallgraph

def main():


    # pycallgraph.start_trace()
    config = GetConfig()
    runConfig( config)
    # pycallgraph.make_dot_graph('test.png')


def runConfig( config ):
    SSet = SequenceSet.SequenceSet( config.inputFName )

    # Select only sequences in  config.inputSeqSubSetFName file
    if  ( config.inputSeqSubSet != [] ) :
        SSet = SSet.extractSubSet( config.inputSeqSubSet )
        
        #
        # Mandatory for the suffix tree
        #

        # Remove directories
        filename = SSet.filename.split( os.sep )[-1]
        SSet.filename = filename
        SSet.writeFastaFile( filename, sep='' )
        config.TmpFiles.append( filename )
        
    # ??? sites
    # File number (used for manipulating exact or 
    # mismatch decoding files)

    if( config.modeExact ) :

        # Create the EC builder
        ECBuilder = BuildEquivClassesLayers.BuildExactEquivClasses( 
                       config.order, mode = 'NLDEquivClasses')

        ec_layers = ECBuilder.build( SSet )

        # Note : the same in compact form  
        #   ec_layers = BuildEquivClassesLayers.BuildExactEquivClasses( 
        #               config.order, mode = 'NLDEquivClasses').build( SSet)

        # Create the Anchor builder
        b_anchors = BuildAnchorSet.BuildAnchorSetFromEClayers( )
        anchors   = b_anchors.build( ec_layers, config.kappa )
        ## Not used
        #    anchors   = b_anchors.buildMotif( ec_layers, len( SSet.SeqNames ) )
        ## Invalid b_anchors.checkPTree( anchors )

        # Filtering anchors
        for smin in range( config.sMin[0], config.sMin[1]+1):

            f_anchors = AnchorOperator.AnchorSminFilter().operate( anchors,  smin )

            #
            #  Outputs
            #
            if( config.outputsFasta or config.outputsAll ) :
                str_ = config.outputsFName
                if( not config.defaultGenericNames ):
                    str_ = str_ % str(smin)
                f_anchors.writeAnchors( str_, SeqSet=SSet )
            if( config.outputswindow or config.outputsAll ) :
                str_ = config.outputsWName
                if( not config.defaultGenericNames ):
                    str_ = str_ % str(smin)
                f_anchors.writeWindows( str_, SeqSet=SSet )
            if( config.outputsWindow or config.outputsAll ) :
                str_ = config.outputsWName
                if( not config.defaultGenericNames ):
                    str_ = str_ % str(smin)
                f_anchors.writeWindows( str_, SeqSet=SSet, PosMode=True )
            if( config.outputsDistance or config.outputsAll ) :
                str_ = config.outputsDName
                if( not config.defaultGenericNames ):
                    str_ = str_ % str(smin)
                dist_matrix = AnchorOperator.DistanceMatrix().operate( 
                    f_anchors, SSet)
                dist_matrix.writeNexus( str_, SSet.SeqNames )

        if( config.outputsPTree or config.outputsAll ) :
            b_anchors.writePTree( config.outputsPName, config.kappa )


    if( config.modeMismatch ) :

        ECBuilder = BuildEquivClassesLayers.BuildMismatchEquivClasses( config.order ) 
        ec_layers = ECBuilder.build( SSet )

        # Create the Anchor builder
        b_anchors = BuildAnchorSet.BuildAnchorSetFromEClayers( )
        anchors   = b_anchors.build( ec_layers, config.kappa )

        # Filtering anchors
        for smin in range( config.sMin[0], config.sMin[1]+1):

            f_anchors = AnchorOperator.AnchorSminFilter().operate( anchors,  smin )

            #
            #  Outputs
            #
            
            if( config.outputsFasta ) :
                str_ = config.outputsFName
                if( not config.defaultGenericNames ):
                    str_ = str_ % str(smin)
                f_anchors.writeAnchors( str_, SeqSet=SSet )
            if( config.outputsWindow ) :
                str_ = config.outputsWName
                if( not config.defaultGenericNames ):
                    str_ = str_ % str(smin)
                f_anchors.writeWindows( str_, SeqSet=SSet )
            if( config.outputsDistance ) :
                str_ = config.outputsDName
                if( not config.defaultGenericNames ):
                    str_ = str_ % str(smin)
                dist_matrix = AnchorOperator.DistanceMatrix().operate( 
                    f_anchors, SSet)
                dist_matrix.writeNexus( str_, SSet.SeqNames )

        if( config.outputsPTree ) :
            b_anchors.writePTree( config.outputsPName, config.kappa )

    # Remove tempory files
    # ??? To uncomment (debug)
    # for i in config.TmpFiles :
    #    os.remove( i )
        
    sys.stderr.write('Fin \n')                                                


if __name__ == "__main__":
    main()
